# Advanced performance evaluator demo

from definitions import *
from pyopus.evaluator.performance import PerformanceEvaluator
from pyopus.design.sensitivity import Sensitivity, screen
from pyopus.parallel.mpi import MPI
from pyopus.parallel.cooperative import cOS
from pprint import pprint
import numpy as np

if __name__=='__main__':
	# Prepare statistical parameters dictionary with nominal values
	nominalStat={ name: desc['dist'].mean() for name, desc in statParams.items() }
	
	# Prepare operating parameters dictionary with nominal values
	nominalOp={ name: desc['init'] for name, desc in opParams.items() }
	
	# Prepare initial design parameters dictionary
	initialDesign={ name: desc['init'] for name, desc in designParams.items() }
	
	designParamNames=sorted(list(initialDesign.keys()))
	
	nominalParams={}
	nominalParams.update(nominalStat)
	nominalParams.update(nominalOp)
	
	# Prepare one corner
	#  defined for all simulator setups
	#  module 'mc' definines the typical Monte Carlo model
	#  operating parameters are set to nominal values
	#  statistical parameters are set to 
	corners={
		'nom': {
			'params': nominalParams, 
			'modules': ['mc']
		}
	}
	
	# Prepare parallel evaluation environment, cOS will use MPI as its backend
	# Remote tasks will be started in their own local folders. Before a task 
	# is started all files in the current folder of the machine that spawns a 
	# the remote task will be copied to that task's local folder (this is 
	# specified by the mirrorMap argument to MPI constructor). 
	cOS.setVM(MPI(mirrorMap={'*':'.'}))
	
	# In definitions.py measures have no corners listed, 
	# Therefore they will be evaluated across all specified corners. 
	# The storeResults argument makes the evaluator store the waveforms
	# in temporary result files in the system TMP folder (because resultsFolder
	# is set to None) with prefix 'restemp_'
	pe=PerformanceEvaluator(
		heads, analyses, measures, corners, variables=variables, 
		activeMeasures=['vgs_drv', 'cmrr'], 
		debug=0
	)
	
	# Sensitivity setups
	sens=Sensitivity(pe, designParams, designParamNames, debug=2, relPerturbHiLo=0.05)
	
	# Compute deltas
	diffs, deltas, anCount = sens(initialDesign)
	
	
	# Extract deltas for cmrr
	dv=sens.diffVector('cmrr', 'nom')
	
	# Parameter screening for cmrr
	print("\nSignificant design parameters influencing cmrr")
	outNdx, inNdx = screen(
		dv, deltas, designParamNames, 
		contribThreshold=0.01, cumulativeThreshold=0.25, useSens=False, squared=True, 
		debug=2
	)
	print([designParamNames[ii] for ii in inNdx])
	
	
	# Extract deltas for vgs_drv
	dv=sens.diffVector('vgs_drv', 'nom')
	
	# Screening for first component of vgs_drv
	print("\nSignificant design parameters influencing vgs_drv[0]")
	outNdx, inNdx = screen(
		dv[:,0], deltas, designParamNames, 
		contribThreshold=0.01, cumulativeThreshold=0.25, useSens=False, squared=True, 
		debug=2
	)
	print([designParamNames[ii] for ii in inNdx])
	
	
	# Prepare for evaluation sensitivities of cmrr to statistical parameters
	nominalParams={}
	nominalParams.update(nominalOp)
	nominalParams.update(initialDesign)
	corners={
		'nom': {
			'params': nominalParams, 
			'modules': ['mc']
		}
	}
	# To make evaluations faster, evaluate only cmrr
	pe=PerformanceEvaluator(
		heads, analyses, measures, corners, variables=variables, 
		activeMeasures=['cmrr'], 
		debug=0
	)
	sensParams={
		name: {
			'lo': -10, 
			'hi': 10, 
			'init': desc['dist'].mean(), 
		} for name, desc in statParams.items()
	}
	sensNames=list(sensParams.keys())
	sens=Sensitivity(pe, sensParams, sensNames, debug=0, relPerturbHiLo=1/64, diffType="central")
	diffs, deltas, anCount = sens()
	dv=sens.diffVector('cmrr', 'nom')
	# Parameter screening for cmrr
	print("\nSignificant statistical parameters influencing cmrr")
	outNdx, inNdx = screen(
		dv, deltas, sensNames, 
		contribThreshold=0.01, cumulativeThreshold=0.25, useSens=False, squared=True, 
		debug=2
	)
	print([sensNames[ii] for ii in inNdx])
	
	
	# Cleanup temporary files generated by the evaluator
	pe.finalize()
	
	# Finalize cOS parallel environment
	cOS.finalize()
	
